import sys
import asyncio
import signal
import functools
from typing import Any, Type

from .constants import *
from .registry import PTRegistry, TokenRegistry
from .pt_common import PTCommon, Arc
from .observer import MergedRecordsType, Observer, ComparativeObserver
from .transition import Transition
from .place import Place, SpecialPlace
from .token import Token


def _int_handler(
    signame: str, loop: asyncio.AbstractEventLoop, pt_registry: PTRegistry
) -> None:
    print(f"Got signal '{signame}'")

    if signame == "SIGINT" or signame == "SIGTERM":
        print(">>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<")
        loop.stop()


def _add_int_handlers(pt_registry: PTRegistry) -> None:
    loop = asyncio.get_running_loop()

    for signame in {"SIGINT", "SIGTERM"}:
        loop.add_signal_handler(
            getattr(signal, signame),
            functools.partial(_int_handler, signame, loop, pt_registry),
        )


def _cancel_all_tasks() -> None:
    tasks: set[asyncio.Task[Any]] = asyncio.all_tasks()
    for task in tasks:
        task.cancel()


def terminate() -> None:
    """
    Terminates PT net simulation.
    """
    _cancel_all_tasks()


async def main(pt_registry: PTRegistry) -> None:
    """
    Main entry point of PT net simulation.

    Runs the tasks assigned to places and transitions registered in ``pt_registry``.

    :param pt_registry: Registry object keeping all places and transitions in the model.
    :param debug_level: Set debug level to ``"ERROR", "DEBUG"`` or ``"INFO"``.
    """
    tasks: set[asyncio.Task[PTCommon]] = set()

    _add_int_handlers(pt_registry)
    for loop in pt_registry.get_loops():
        task: asyncio.Task[Any] = asyncio.create_task(loop)
        tasks.add(task)
        task.add_done_callback(tasks.discard)

    await asyncio.gather(*tasks, return_exceptions=False)


class SoyutNet(object):
    def __init__(self) -> None:
        self._LOOP_DELAY: float = 0.5
        self.DEBUG_ENABLED: bool = False
        """if set, :py:func:`soyutnet.SoyutNet.DEBUG` will print."""
        self.VERBOSE_ENABLED: bool = False
        """if set, :py:func:`soyutnet.SoyutNet.DEBUG_V` will print."""
        self.SLOW_MOTION: bool = False
        """If set, task loops are delayed for :py:attr:`soyutnet.SoyutNet.LOOP_DELAY` seconds"""

    @property
    def LOOP_DELAY(self) -> float:
        """
        Asyncio tasks main loop delay for debugging.

        :return: Delay amount in seconds.
        """
        if self.SLOW_MOTION:
            return self._LOOP_DELAY
        return 0

    async def sleep(self, amount: float = 0.0) -> None:
        """
        Wrapper for task sleep function.

        :param amount: Sleep amount in seconds.
        """
        await asyncio.sleep(amount)

    def time(self) -> float:
        """
        Get current time since the program starts.

        :return: Current time in seconds.
        """
        loop = asyncio.get_running_loop()
        return loop.time()

    def get_loop_name(self) -> str:
        """
        Get the name of current loop from which this function is called.

        :return: Name of the loop.
        """
        name: str = "NO-LOOP"
        try:
            task: asyncio.Task[Any] | None = asyncio.current_task()
            if isinstance(task, asyncio.Task):
                name = task.get_name()
        except RuntimeError:
            pass

        return name

    def DEBUG_V(self, *args: Any) -> None:
        """
        Print debug messages when :py:attr:`soyutnet.SoyutNet.VERBOSE_ENABLED`.
        """
        if self.DEBUG_ENABLED and self.VERBOSE_ENABLED:
            print(f"{self.get_loop_name()}:", *args)

    def ERROR_V(self, *args: Any) -> None:
        """
        Print error messages when :py:attr:`soyutnet.SoyutNet.VERBOSE_ENABLED`.
        """
        if self.VERBOSE_ENABLED:
            print(f"{self.get_loop_name()}:", *args, file=sys.stderr)

    def DEBUG(self, *args: Any) -> None:
        """
        Print debug messages when :py:attr:`soyutnet.SoyutNet.DEBUG_ENABLED`.
        """
        if self.DEBUG_ENABLED:
            print(f"{self.get_loop_name()}:", *args)

    def ERROR(self, *args: Any) -> None:
        """
        Print error messages.
        """
        print(f"{self.get_loop_name()}:", *args, file=sys.stderr)

    def Arc(self, *args: Any, **kwargs: Any) -> Arc:
        kwargs["net"] = self
        return Arc(*args, **kwargs)

    def Token(self, *args: Any, **kwargs: Any) -> Token:
        kwargs["net"] = self
        return Token(*args, **kwargs)

    def Place(self, *args: Any, **kwargs: Any) -> Place:
        kwargs["net"] = self
        return Place(*args, **kwargs)

    def SpecialPlace(self, *args: Any, **kwargs: Any) -> SpecialPlace:
        kwargs["net"] = self
        return SpecialPlace(*args, **kwargs)

    def Transition(self, *args: Any, **kwargs: Any) -> Transition:
        kwargs["net"] = self
        return Transition(*args, **kwargs)

    def Observer(self, *args: Any, **kwargs: Any) -> Observer:
        kwargs["net"] = self
        return Observer(*args, **kwargs)

    def ComparativeObserver(self, *args: Any, **kwargs: Any) -> ComparativeObserver:
        kwargs["net"] = self
        return ComparativeObserver(*args, **kwargs)

    def TokenRegistry(self, *args: Any, **kwargs: Any) -> TokenRegistry:
        kwargs["net"] = self
        return TokenRegistry(*args, **kwargs)

    def PTRegistry(self, *args: Any, **kwargs: Any) -> PTRegistry:
        kwargs["net"] = self
        return PTRegistry(*args, **kwargs)



import weakref
from weakref import ReferenceType
from typing import (
    Tuple,
    Never,
    Dict,
    TYPE_CHECKING,
)

label_t = int
"""Label type"""
id_t = int
"""ID type"""
TokenType = Tuple[label_t, id_t] | Tuple[Never, ...]
"""Token type"""
TokenWalletType = Dict[label_t, list[id_t]]
"""Token container type"""

if TYPE_CHECKING:
    from . import SoyutNet

INVALID_LABEL: label_t = -10
"""Invalid label"""
INVALID_ID: id_t = -11
"""Invalid id"""
GENERIC_LABEL: label_t = 0
"""Generic label"""
GENERIC_ID: id_t = 0
"""Generic ID"""
INITIAL_ID: id_t = 0


class BaseObject(object):
    """
    Base SoyutNet object inherited by all classes.
    """

    def __init__(self, net: "SoyutNet") -> None:
        self._net: ReferenceType["SoyutNet"] = weakref.ref(net)
        self._ident: str = ""
        """Every object should have a unique string identifier."""

    @property
    def net(self) -> "SoyutNet":
        """
        Reference of the SoyutNet instance assigned to instances of all object types.
        """
        return self._net()  # type: ignore[return-value]


class SoyutNetError(Exception):
    """
    Generic error class left as future work.
    """

    def __init__(self, message: str = "An error occured.") -> None:
        self.message: str = message
        super().__init__(self.message)



import sys
import asyncio

import soyutnet
from soyutnet import SoyutNet
from soyutnet.constants import GENERIC_ID, GENERIC_LABEL


def main(w1=1, w2=1):
    token_ids = [GENERIC_ID] * 35

    async def producer(place):
        try:
            id: id_t = token_ids.pop(0)
            return (GENERIC_LABEL, id)
        except IndexError:
            pass

        return tuple()

    async def consumer(place):
        return

    place_count = 3

    def on_comparison_ends(observer):
        nonlocal place_count
        place_count -= 1
        print(place_count)
        if place_count == 0:
            soyutnet.terminate()

    w3 = w1 + w2

    net = soyutnet.SoyutNet()

    o00 = net.Observer(verbose=False)
    o01 = net.Observer(verbose=False)
    o10 = net.ComparativeObserver(
        expected={1: [((GENERIC_LABEL, 3),)] * 3},
        on_comparison_ends=on_comparison_ends,
        verbose=False,
    )
    o11 = net.ComparativeObserver(
        expected={1: [((GENERIC_LABEL, 3),)] * 3},
        on_comparison_ends=on_comparison_ends,
        verbose=False,
    )
    o2 = net.ComparativeObserver(
        expected={
            1: [
                ((GENERIC_LABEL, i),)
                for i in [5, 4, 3, 7, 6, 5, 9, 8, 7, 11, 10, 9, 13, 12, 11]
                + list(range(15, 1, -1))
            ]
        },
        on_comparison_ends=on_comparison_ends,
        verbose=False,
    )
    o3 = net.Observer(verbose=False)

    p00 = net.SpecialPlace("p00", producer=producer, observer=o00)
    p01 = net.SpecialPlace("p01", producer=producer, observer=o01)
    p10 = net.Place("p10", observer=o10)
    p11 = net.Place("p11", observer=o11)
    p2 = net.Place("p2", observer=o2)
    p3 = net.SpecialPlace("p3", consumer=consumer, observer=o3)

    t00 = net.Transition("t00")
    t01 = net.Transition("t01")
    t1 = net.Transition("t1")
    t2 = net.Transition("t2")

    reg = net.PTRegistry()

    reg.register(p00)
    reg.register(p01)
    reg.register(p10)
    reg.register(p11)
    reg.register(p2)
    reg.register(p3)

    reg.register(t00)
    reg.register(t01)
    reg.register(t1)
    reg.register(t2)

    (
        (
            p00.connect(t00).connect(p10).connect(t1, weight=w1),
            p01.connect(t01).connect(p11).connect(t1, weight=w2),
        )[0]
        .connect(p2, weight=w3)
        .connect(t2, weight=1)
        .connect(p3, weight=1)
    )

    try:
        asyncio.run(soyutnet.main(reg))
    except asyncio.exceptions.CancelledError:
        print("Simulation is terminated.")

    gv = reg.generate_graph()
    return gv


if __name__ == "__main__":
    gv = main(int(sys.argv[1]), int(sys.argv[2]))
    with open("test.gv", "w") as fh:
        fh.write(gv)



import sys
import asyncio
from weakref import ref, ReferenceType
from typing import (
    Any,
    AsyncGenerator,
    Self,
    Dict,
    Tuple,
    Generator,
    Awaitable,
    Callable,
    TYPE_CHECKING,
)

from .constants import *
from .token import Token
from .observer import Observer


if TYPE_CHECKING:
    Queue = asyncio.Queue[Any]
else:
    Queue = asyncio.Queue


class Arc(object):
    """
    Defines a generic labeled PT net arc which connects places to transitions or vice versa.
    """

    def __init__(
        self,
        start: Any,
        end: Any,
        weight: int = 1,
        labels: list[label_t] = [GENERIC_LABEL],
    ) -> None:
        """
        Constructor.

        :param start: Place or transition. Input place of a transition (`end`), or input transition of a place (`end`).
        :param end: Place or transition. Output place of a transition (`start`), or output transition of a place (`start`).
        :param weight: Arc weight.
        :param labels: List of arc label.
        """
        self.start: ReferenceType[Any] = ref(start)
        """Input place/transition"""
        self.index_at_start: int = -1
        """Index in the list of output arcs of :py:attr:`soyutnet.pt_common.Arc.start`"""
        self.end: ReferenceType[Any] = ref(end)
        """Output place/transition"""
        self.index_at_end: int = -1
        """Index in the list of output arcs of :py:attr:`soyutnet.pt_common.Arc.end`"""
        self.weight: int = weight
        """Arc weight"""
        self.labels: list[label_t] = list(labels)
        """The list of arc labels"""
        self._queue: Queue = Queue(maxsize=weight)
        """Input/output queue for transmitting tokens from :py:attr:`soyutnet.pt_common.Arc.start` to :py:attr:`soyutnet.pt_common.Arc.end`"""

    def __str__(self) -> str:
        """
        Returns string representation of the arc.

        :return: String representation of the arc.
        """
        start_ident: str = ""
        start_ref: Any = self.start()
        if start_ref is not None:
            start_ident = start_ref.ident()
        end_ident: str = ""
        end_ref: Any = self.end()
        if end_ref is not None:
            end_ident = end_ref.ident()
        return (
            f"{start_ident}:{self.index_at_start} -> {end_ident}:{self.index_at_end}, "
            f"l={self.labels}, w={self.weight}"
        )

    async def wait(self) -> AsyncGenerator[TokenType, None]:
        """
        Acquires :py:attr:`soyutnet.pt_common.Arc.weight` tokens from :py:attr:`soyutnet.pt_common.Arc.start` and yields them to :py:attr:`soyutnet.pt_common.Arc.end`

        :return: Tokens.
        """
        count: int = self.weight
        while count > 0:
            token: TokenType = await self._queue.get()
            self._queue.task_done()
            count -= 1
            yield token

    async def send(self, token: TokenType) -> None:
        """
        Puts a token to the output arc.

        :param token: Token.
        """
        if not token:
            return
        await self._queue.put(token)

    def is_enabled(self) -> bool:
        """
        It is checked by the output transition at :py:attr:`soyutnet.pt_common.Arc.end` to detemine the transition is enabled or not.

        :return: ``True`` if enabled.
        """
        return self._queue.full()

    async def observe_input_places(self, requester: str = "") -> None:
        """
        It is called by the output transition at :py:attr:`soyutnet.pt_common.Arc.end` after the transition is enabled.

        It records the tokens counts just before the transition happens.

        :param requester: The identity of caller.
        """
        start_ref: Any = self.start()
        if start_ref is not None:
            await start_ref.observe(requester=requester)

    async def notify_observer(self, label: label_t, increment: int = -1) -> None:
        start_ref: Any = self.start()
        if start_ref is not None:
            await start_ref._observer.inc_token_count(label, increment)

    def get_graphviz_definition(self, t: int) -> str:
        start_ref: Any = self.start()
        end_ref: Any = self.end()
        if end_ref is not None and start_ref is not None:
            return f"""{start_ref._name}_{t} -> {end_ref._name}_{t} [fontsize="20",label=" {self.weight}",minlen="2",penwidth="3"];"""

        return ""


class PTCommon(Token):
    """
    Base class implementing shared properties of places and transitions.
    """

    def __init__(
        self,
        name: str = "",
        initial_tokens: TokenWalletType = {},
        processor: Callable[["PTCommon"], Awaitable[bool]] | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Constructor.

        :param name: Name of the place or transition.
        :param initial_tokens: Dictionary of initial tokens, in other words initial marking of the place.
        :param processor: Custom token processing function that is called between processing input and output arcs.
        """
        super().__init__(**kwargs)
        self._name: str = name
        """Name of the PT"""
        self._input_arcs: list[Arc] = []
        """List of input arcs"""
        self._last_processed_input_arc_index: int = 0
        self._output_arcs: list[Arc] = []
        """List of output arcs"""
        self._last_processed_output_arc_index: int = 0
        self._tokens: TokenWalletType = {
            label: list(initial_tokens[label]) for label in initial_tokens
        }
        """Keeps tokens"""
        self._observer: Observer | None = None
        """Observes the tokens before each firing of output transitions"""
        self._ident: str = self.ident()
        """Unique identifier"""
        self._processor: Callable[["PTCommon"], Awaitable[bool]] | None = processor
        """Custom token processing function that is called between processing input and output arcs"""

    def _put_token(self, token: TokenType, strict: bool = True) -> int:
        """
        Places tokens into a list based on its label.

        :param token: A label and ID pair.
        :param strict: If set the label of token must already be in :py:attr:`soyutnet.pt_common.PTCommon._tokens` dictionary.
        :return: Number of tokens with the given label.
        """
        label: label_t = token[0]
        id: id_t = token[1]
        """NOTE: ``connect`` must add an item with key = ``label`` to the ``self._tokens`` dict."""
        if not strict and label not in self._tokens:
            self._tokens[label] = []
        try:
            self._tokens[label].append(id)
        except KeyError as e:
            # TODO: Handle model error
            _, _, exc_tb = sys.exc_info()
            if exc_tb is not None:
                raise RuntimeError(
                    f"{self.ident()}: {self._tokens} {label} {e} [{exc_tb.tb_frame}, {exc_tb.tb_lineno}, {exc_tb.tb_lasti}]"
                )
            raise KeyError(e)

        return self._get_token_count(label)

    def _get_token(self, label: label_t) -> TokenType:
        """
        Gets the first token with the given label from a FIFO list.

        :param label: Label.
        :return: Token.
        """
        try:
            id: id_t = self._tokens[label].pop(0)
            return (label, id)
        except KeyError as e:
            """Raised when label is not in ``self._tokens``."""
            # TODO: Handle model error
            _, _, exc_tb = sys.exc_info()
            if exc_tb is not None:
                self.net.ERROR_V(
                    f"{e} [{exc_tb.tb_frame}, {exc_tb.tb_lineno}, {exc_tb.tb_lasti}]"
                )
        except IndexError as e:
            """Raised when no token left with ``label`` in ``self._tokens``."""
            # TODO: Is it a model error or can be passed?
            _, _, exc_tb = sys.exc_info()
            if exc_tb is not None:
                self.net.ERROR_V(
                    f"{e} [{exc_tb.tb_frame}, {exc_tb.tb_lineno}, {exc_tb.tb_lasti}]"
                )

        return tuple()

    def _get_token_count(self, label: label_t) -> int:
        """
        Get the number of tokens with the given label.

        :param label: Label.
        :return: Number of tokens with the given label.
        """
        # TODO: Handle model error
        return len(self._tokens[label])

    async def _get_input_arcs(self) -> AsyncGenerator[Arc, None]:
        """
        Generator to iterate through input arcs.

        :return: Input arcs.
        """
        count: int = len(self._input_arcs)
        i: int = 0
        while i < count:
            yield self._input_arcs[self._last_processed_input_arc_index]
            self._last_processed_input_arc_index += 1
            if self._last_processed_input_arc_index == count:
                self._last_processed_input_arc_index = 0
            i += 1

    async def _get_output_arcs(self) -> AsyncGenerator[Arc, None]:
        """
        Generator to iterate through output arcs.

        :return: Output arcs.
        """
        count: int = len(self._output_arcs)
        i: int = 0
        while i < count:
            yield self._output_arcs[self._last_processed_output_arc_index]
            self._last_processed_output_arc_index += 1
            if self._last_processed_output_arc_index == count:
                self._last_processed_output_arc_index = 0
            i += 1

    async def _process_input_arcs(self) -> bool:
        """
        Acquires tokens from enabled input arcs and stores them.

        :return: If ``True`` proceeds to processing tokens and output arcs, else continues waiting for enabled arcs.
        """
        self.net.DEBUG_V(f"{self._ident}: process_input_arcs")
        async for arc in self._get_input_arcs():
            if not arc.is_enabled():
                self.net.DEBUG_V(f"Not enabled {arc}")
                continue
            async for token in arc.wait():
                self.net.DEBUG_V(f"Received '{token}' from {arc}")
                self._put_token(token)
                if self._observer is not None:
                    await self._observer.inc_token_count(token[0])

        return True

    async def _process_output_arcs(self) -> None:
        """
        Sends tokens to the output PTs.
        """
        self.net.DEBUG_V(f"{self._ident}: process_output_arcs")
        async for arc in self._get_output_arcs():
            token: TokenType = tuple()
            for label in arc.labels:
                token = self._get_token(label)
                if token:
                    break
            if not token:
                self.net.DEBUG_V(f"No token, skipping '{arc}'")
                continue
            self.net.DEBUG_V(f"Sending '{token}' to {arc}")
            await arc.send(token)

    async def _process_tokens(self) -> bool:
        """
        Processes input tokens before sending if required.

        :return: ``True`` by default, else goes back to :py:func:`soyutnet.pt_common.PTCommon._process_input_arcs`.
        """
        self.net.DEBUG_V(f"{self._ident}: process_tokens")
        if self._processor is None:
            return True

        return await self._processor(self)

    async def _observe(self, requester: str = "") -> None:
        """
        Dummy observer.
        """
        pass

    async def _set_initial_marking(self) -> None:
        if self._observer is not None:
            for label in self._tokens:
                await self._observer.inc_token_count(
                    label, self._get_token_count(label)
                )

    def ident(self) -> str:
        """
        Returns the unique identifier of PT.

        :return: Unique identifier.
        """
        return f"({self._name}, {self._id})"

    async def should_continue(self) -> bool:
        """
        Main loop of async task assigned to the PT.

        :return: Continues task if `True`.
        """
        if not await self._process_input_arcs():
            return True

        if not await self._process_tokens():
            """If ``False`` do not process output arcs yet."""
            return True

        await self._process_output_arcs()

        return True

    def connect(
        self, other: Self, weight: int = 1, labels: list[label_t] = [GENERIC_LABEL]
    ) -> Self:
        """
        Connects the output of `self` to the input of an other PT by creating an Arc in between.

        :param other: The place/transition which it will be connected to.
        :param weight: Arc weight.
        :param labels: List of arc labels.
        :return: Output place or transition that ``other`` references.
        """
        arc: Arc = Arc(start=self, end=other, weight=weight, labels=list(labels))
        self._output_arcs.append(arc)
        other._input_arcs.append(arc)
        arc.index_at_start = len(self._output_arcs) - 1
        arc.index_at_end = len(other._input_arcs) - 1
        for label in arc.labels:
            if label not in self._tokens:
                self._tokens[label] = []
            if label not in other._tokens:
                other._tokens[label] = []
        self.net.DEBUG_V(f"{self._ident}: Connected arc: {str(arc)}")

        return other

    async def observe(self, requester: str = "") -> None:
        """
        Public observe function called by output arcs.

        :param requester: The identity of caller.
        """
        await self._observe(requester=requester)

    def put_token(self, label: label_t = GENERIC_LABEL, id: id_t = GENERIC_ID) -> int:
        """
        Places tokens into a list based on its label.

        :param label: Label.
        :param id: ID.
        :return: Number of tokens with the given label.
        """
        return self._put_token((label, id))

    def get_token(self, label: label_t) -> TokenType:
        """
        Gets the first token with the given label from a FIFO list.

        :param label: Label.
        :return: A token if exists else empty token which is a null tuple (``tuple()``).
        """
        return self._get_token(label)

    def get_token_count(self, label: label_t) -> int:
        """
        Get the number of tokens with the given label.

        :param label: Label.
        :return: Number of tokens.
        """
        return self._get_token_count(label)


async def _loop(pt: PTCommon) -> None:
    """
    Task function assigned to the PT.

    :param pt: PTCommon instance.
    """
    task: asyncio.Task[Any] | None = asyncio.current_task()
    if isinstance(task, asyncio.Task):
        task.set_name(f"loop{pt.ident()}")
    else:
        return

    await pt._set_initial_marking()
    pt.net.DEBUG_V(f"{pt.ident()}: Loop started")

    while await pt.should_continue():
        await pt.net.sleep(pt.net.LOOP_DELAY)
        pass

    pt.net.DEBUG_V(f"{pt.ident()}: Loop ended")



import os
import asyncio
from typing import (
    Any,
    Dict,
    Callable,
    Awaitable,
    Generator,
    Tuple,
    Coroutine,
)

from .constants import *
from .pt_common import PTCommon, _loop
from .observer import Observer, ObserverRecordType, MergedRecordsType
from .token import Token
from .place import Place, SpecialPlace
from .transition import Transition

DirectoryType = Dict[label_t, list[Tuple[id_t, Any]]]
"""Registry directory type"""
PostRegisterCallbackType = Callable[[id_t, Any], None]
"""Type of callbacks run after an object is registered"""


def _default_post_register_callback(dummy1: Any, dummy2: int) -> None:
    """Default dummy :py:attr:`soyutnet.registry.PostRegisterCallbackType`"""
    pass


class Registry(BaseObject):
    """
    Registry keeps track of (label, id) tuples and the objects assigned to them.
    It generates unique ids for new objects.
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._id_counter: id_t = INITIAL_ID
        """Auto-incrementing id assigned to new objects"""
        self._directory: DirectoryType = {}
        """Keeps all objects categorized by labels"""
        self._lock: asyncio.Lock = asyncio.Lock()
        """Locks access to :py:attr:`soyutnet.registry.Registry._directory`"""

    def _new_id(self) -> id_t:
        """
        Creates new ids for new objects.

        :return: Unique id
        """
        self._id_counter += 1
        return self._id_counter

    def register(
        self,
        obj: Any,
        post_register_callback: PostRegisterCallbackType = _default_post_register_callback,
    ) -> id_t:
        """
        Register a new object

        :param obj: New object of any type.
        :param post_register_callback: Called after object is registered.
        :return: Assigned unique ID.
        """
        new_id: id_t = self._new_id()
        label: label_t = obj.get_label()
        if label not in self._directory:
            self._directory[label] = []
        self._directory[label].append((new_id, obj))
        if post_register_callback != _default_post_register_callback:
            post_register_callback(new_id, obj)

        return new_id

    def get_entry_count(self, label: label_t = GENERIC_LABEL) -> int:
        """
        Returns the number of entries with the given label.

        :param label: Label.
        :return: Number of entries.
        """
        if label in self._directory:
            return len(self._directory)

        return 0

    def get_first_entry(self, label: label_t = GENERIC_LABEL) -> Tuple[id_t, Any]:
        """
        Returns first entry with given label. First entry is the one registered first.

        :param label: Label.
        :return: Entry.
        """
        if label in self._directory and len(self._directory[label]) > 0:
            return self._directory[label][0]

        return (INVALID_ID, None)

    def get_entries(self, label: label_t, id: id_t) -> list[Any]:
        """
        Returns a list of objects with given label and ID.

        :param label: Label.
        :param id: ID.
        :return: A list of objects.
        """
        result: list[Any] = []
        if label not in self._directory:
            return result

        for entry in self._directory[label]:
            if entry[0] == id:
                result.append(entry[1])

        return result

    def entries(self, label: label_t | None = None) -> Generator[Any, None, None]:
        """
        Iterates through all entries with the given label.

        :param label: Label. If it is ``None``, iterates through all labels.
        :return: Yields entries
        """
        d: DirectoryType = (
            {label: self._directory[label]}
            if label in self._directory
            else self._directory
        )
        for label in d:
            for entry in d[label]:
                yield entry


class TokenRegistry(Registry):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    def register(self, token: Token) -> id_t:  # type: ignore[override]
        """
        Register a new token

        :param token: New token.
        :return: Assigned unique ID.
        """

        def callback(new_id: id_t, tkn: Any) -> None:

            tkn._id = new_id

        return super().register(token, callback)


class PTRegistry(Registry):
    """
    Keeps track of PTCommon instances.
    """

    def __init__(self, **kwargs: Any) -> None:
        """
        Constructor.
        """
        super().__init__(**kwargs)

    def get_loops(self) -> Generator[Coroutine[Any, Any, None], None, None]:
        """
        Yields asyncio task functions assigned to the PT.

        :return: Asyncio task function.
        """
        for label in self._directory:
            for entry in self._directory[label]:
                yield _loop(entry[1])

    def register(self, pt: PTCommon) -> id_t:  # type: ignore[override]
        """
        Registers a PT.

        :param pt: PTCommon instance.
        :return: Unique ID assigned to the PT.
        """

        def callback(new_id: id_t, pt: PTCommon) -> None:
            pt._id = new_id
            if not pt._name:
                class_name: str = type(pt).__name__
                if isinstance(pt, Place):
                    pt._name = f"p{pt._id}"
                elif isinstance(pt, Transition):
                    pt._name = f"t{pt._id}"
            self.net.DEBUG_V(f"Registered: {pt.ident()}")

        return super().register(pt, callback)

    def get_merged_records(
        self, ignore_special_places: bool = True
    ) -> MergedRecordsType:
        """
        Merges all observer records and sorts by their timestamps.

        :return: Merged and sorted observer records.
        """
        output: list[Tuple[str, ObserverRecordType | Tuple[float]]] = []
        for e in self.entries():
            obj: Any = e[1]
            if not isinstance(obj, PTCommon):
                continue
            if ignore_special_places and isinstance(obj, SpecialPlace):
                continue
            name: str = obj._name
            if obj._observer is None:
                continue
            obsv: Observer = obj._observer
            for record in obsv.get_records():
                output.append((name, record))

        return output

    def _get_graphviz_node_definition(self, pt: PTCommon, t: int) -> str:
        shape: str = "circle"
        color: str = "#000000"
        fillcolor: str = "#dddddd"
        height: float = 1
        width: float = 1
        fontsize: int = 20
        penwidth: int = 3
        if isinstance(pt, Transition):
            shape = "box"
            color = "#cccccc"
            fillcolor = "#000000"
            height = 0.25
            width = 1.25
        node_fmt: str = (
            """{}_{} [shape="{}",fontsize="{}",style="filled",color="{}",fillcolor="{}",label="",xlabel="{}",height="{}",width="{}",penwidth={}];"""
        )

        return node_fmt.format(
            pt._name,
            t,
            shape,
            fontsize,
            color,
            fillcolor,
            pt._name,
            height,
            width,
            penwidth,
        )

    def generate_graph(self, net_name: str = "Net") -> str:
        eol: str = os.linesep
        gv: str = f"digraph {net_name} {{" + eol
        gv_nodes: str = ""
        gv_arcs: str = ""
        indent = "\t"
        for e in self.entries():
            obj: Any = e[1]
            if not isinstance(obj, PTCommon):
                continue
            node_def: str = self._get_graphviz_node_definition(obj, 0)
            gv_nodes += 2 * indent + node_def + eol
            for arc in obj._input_arcs:
                gv_arcs += 2 * indent + arc.get_graphviz_definition(0) + eol

        gv += indent + "subgraph cluster_0 {" + eol
        gv += 2 * indent + "penwidth=3;" + eol
        gv += gv_nodes
        gv += gv_arcs
        gv += indent + "}" + eol
        gv += indent + "clusterrank=none;" + eol
        gv += "}" + eol

        return gv



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import sys
import os
sys.path.insert(0, os.path.abspath('../../soyutnet/'))
sys.path.insert(0, os.path.abspath('../../'))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'SoyutNet'
copyright = '2024, Okan Demir'
author = 'Okan Demir'
release = '0.1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
]
autosummary_generate = True
templates_path = ['_templates']
exclude_patterns = []
nitpicky = True



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'classic'
html_static_path = ['_static']
html_logo = "_static/soyutnet_logo.png"
html_theme_options = {
    "footerbgcolor": "#ffffff",
    "bgcolor": "#efefef",
    "sidebarbgcolor": "#f6f6f6",
    "sidebartextcolor": "#000000",
    "sidebarlinkcolor": "#496e89",
    "headbgcolor": "#edf6ff66",
    "stickysidebar": True,
}

autodoc_default_options = {
    "members": True,
    "undoc-members": False,
    "private-members": True,
}



import asyncio
from typing import (
    Any,
    Dict,
)

from .constants import *
from .pt_common import PTCommon

FiringRecordType = Tuple[float]
"""Firing record type"""
FiringHistoryType = list[FiringRecordType]
"""Type for list of firing records"""


class Transition(PTCommon):
    """
    Defines PTNet transitions.
    """

    def __init__(
        self, name: str = "", record_firing: bool = False, **kwargs: Any
    ) -> None:
        """
        Constructor.

        :param name: Name of the transition.
        """
        super().__init__(name=name, **kwargs)
        self._no_of_times_enabled: int = 0
        """Counts the number of time the transition is enabled"""
        self._firing_records: FiringHistoryType = []
        """Keeps timestampts of each firing of the transition: py:attr:`soyutnet.transition.FiringRecordType`"""
        self._record_firing: bool = record_firing
        """Enables recording firings of transitions"""

    def _new_firing_record(self) -> None:
        self._firing_records.append((self.net.time(),))

    async def _process_input_arcs(self) -> bool:
        """
        Acquires and stores tokens.

        :return: ``True`` if the transition is enabled, else goes back to waiting input arcs to be enabled.
        """
        self.net.DEBUG_V(f"{self._ident}: process_input_arcs")
        async for arc in self._get_input_arcs():
            if not arc.is_enabled():
                return False

        self.net.DEBUG_V(f"Enabled!")
        self._no_of_times_enabled += 1
        if self._record_firing:
            self._new_firing_record()

        async for arc in self._get_input_arcs():
            await arc.observe_input_places(self._name)
            count: int = arc.weight
            async for token in arc.wait():
                self.net.DEBUG_V(f"Received '{token}' from {arc}")
                self._put_token(token)
                await arc.notify_observer(token[0])
                count -= 1
                if count <= 0:
                    break

        return True

    async def _process_tokens(self) -> bool:
        """
        Calls ``super()``'s version.

        See, :py:func:`soyutnet.pt_common.PTCommon._process_tokens`.
        """
        return await super()._process_tokens()

    async def _process_output_arcs(self) -> None:
        """
        Fires the transition.

        NOTE: sum of w(p_prev, self) == sum of w(self, p_next) must satisfy for each label.

        Sends tokens to the output places when required conditions are satisfied.
        """
        self.net.DEBUG_V(f"{self._ident}: process_output_arcs")
        async for arc in self._get_output_arcs():
            count: int = arc.weight
            while count > 0:
                token: TokenType = tuple()
                for label in arc.labels:
                    token = self._get_token(label)
                    if token:
                        break
                if not token:
                    break
                self.net.DEBUG_V(f"Sending '{token}' to {arc}")
                await arc.send(token)
                count -= 1

    def get_no_of_times_enabled(self) -> int:
        """
        Returns number of times the transition is enabled.

        :return: Number of times the transition is enabled.
        """
        return self._no_of_times_enabled

    def get_firing_records(self) -> FiringHistoryType:
        """
        Returns all firing records. :py:attr:`soyutnet.transition.Transition._firing_records`

        :return: Firing records.
        """
        return self._firing_records



import pytest
import asyncio

import soyutnet
from soyutnet import SoyutNet
from soyutnet.constants import GENERIC_LABEL, GENERIC_ID, INVALID_ID, INITIAL_ID


@pytest.mark.asyncio
async def test_01():
    net = SoyutNet()
    registry = net.TokenRegistry()
    token = net.Token()

    assert (INVALID_ID, None) == registry.get_first_entry(token.get_label())
    registry.register(token)
    assert token.get_id() == INITIAL_ID + 1
    assert token.get_label() == GENERIC_LABEL
    assert (token.get_id(), token) == registry.get_first_entry(token.get_label())

    assert registry.get_entry_count() == 1


@pytest.mark.asyncio
async def test_02():
    net = SoyutNet()
    place = net.Place()

    assert place.get_id() == GENERIC_ID
    assert place.get_binding() is None


@pytest.mark.asyncio
async def test_03():
    net = SoyutNet()
    transition = net.Transition()

    assert transition.get_id() == GENERIC_ID
    assert transition.get_binding() is None


def test_04():
    import simple_example as e

    e.main()


def test_05():
    import simple_example_different_weight as e

    e.main_01(3, 2)


def test_06():
    import simple_example_two_input_places as e

    for i in range(100, 10000, 1000):
        e.main(i)


def test_07():
    import simple_example_two_input_places_but_different_weights as e

    e.main(w1=3, w2=2)


def test_08():
    from basic_models import co_begin

    for i in range(2, 100):
        co_begin(i)


def test_09():
    from basic_models import co_end

    for i in range(2, 100):
        co_end(i)


def test_10():
    from basic_models import sync_by_signal

    sync_by_signal()


def test_11():
    from n_tester import n_tester

    n_tester()



import sys
import asyncio

import soyutnet
from soyutnet import SoyutNet
from soyutnet.constants import GENERIC_LABEL, GENERIC_ID


def main_01(w1=1, w2=1):
    token_ids = [GENERIC_ID] * (w1 * w2 * 10)

    async def producer(place):
        try:
            id: id_t = token_ids.pop(0)
            return (GENERIC_LABEL, id)
        except IndexError:
            pass

        return tuple()

    async def consumer(place):
        return

    place_count = 3

    def on_comparison_ends(observer):
        nonlocal place_count
        place_count -= 1
        if place_count == 0:
            soyutnet.terminate()

    net = SoyutNet()

    expected0 = {1: [((GENERIC_LABEL, i),) for i in [1, 2, 3, 1, 2, 3, 1, 2, 3]]}
    expected1 = {1: [((GENERIC_LABEL, i),) for i in [3, 4, 2, 3, 4, 2, 3, 4, 2]]}
    expected2 = {
        1: [((GENERIC_LABEL, i),) for i in [0, 0, 0, 0, 0, 2, 2, 4, 4, 6, 6, 8, 8, 10]]
    }

    o0 = net.ComparativeObserver(
        expected=expected0,
        on_comparison_ends=on_comparison_ends,
        verbose=False,
    )
    o1 = net.ComparativeObserver(
        expected=expected1,
        on_comparison_ends=on_comparison_ends,
        verbose=False,
    )
    o2 = net.ComparativeObserver(
        expected=expected2,
        on_comparison_ends=on_comparison_ends,
        verbose=False,
    )

    p0 = net.SpecialPlace("p0", producer=producer, observer=o0)
    p1 = net.Place("p1", observer=o1)
    p2 = net.SpecialPlace("p2", consumer=consumer, observer=o2)
    t1 = net.Transition("t1")
    t2 = net.Transition("t2")

    reg = net.PTRegistry()

    reg.register(p0)
    reg.register(p1)
    reg.register(p2)
    reg.register(t1)
    reg.register(t2)

    p0.connect(t1, weight=w1).connect(p1, weight=w1).connect(t2, weight=w2).connect(
        p2, weight=w2
    )

    try:
        asyncio.run(soyutnet.main(reg))
    except asyncio.exceptions.CancelledError:
        print("Simulation is terminated.")


if __name__ == "__main__":
    main_01(int(sys.argv[1]), int(sys.argv[2]))
